from abc import ABC

from .base import Agent


class HighFrequencyAgent(Agent, ABC):
    """High-frequency trading Agent class.

    High-frequency traders are treated as a special agent. Please see xxxx for more details.
    #ToDo

    This class inherits from the :class:`pams.agents.Agent` class and ABC class.
    """

    pass



import json
import os
import random
import time
from abc import ABC
from abc import abstractmethod
from io import TextIOWrapper
from typing import Dict
from typing import List
from typing import Optional
from typing import Type
from typing import Union

from ..logs.base import Logger
from ..simulator import Simulator


class Runner(ABC):
    """Runner of the market simulator class (Abstract class).

    .. seealso::
        - :class:`pams.runners.SequentialRunner`
    """

    def __init__(
        self,
        settings: Union[Dict, TextIOWrapper, os.PathLike, str],
        prng: Optional[random.Random] = None,
        logger: Optional[Logger] = None,
        simulator_class: Type[Simulator] = Simulator,
    ):
        """initialize.

        Args:
            settings (Union[Dict, TextIOWrapper, os.PathLike, str]): runner configuration.
                You can set python dictionary, a file pointer, or a file path.
            prng (random.Random, Optional): pseudo random number generator for this runner.
            logger (Logger, Optional): logger instance.
            simulator_class (Type[Simulator]): type of simulator.

        Returns:
            None
        """
        self.settings: Dict
        if isinstance(settings, Dict):
            self.settings = settings
        elif isinstance(settings, TextIOWrapper):
            self.settings = json.load(fp=settings)
        else:
            self.settings = json.load(fp=open(settings, mode="r"))
        self._prng: random.Random = prng if prng is not None else random.Random()
        self.logger = logger
        self.simulator: Simulator = simulator_class(
            prng=random.Random(self._prng.randint(0, 2**31))
        )
        self.registered_class: List[Type] = []

    def main(self) -> None:
        """main process. The process is executed while measuring time."""
        setup_start_time_ns = time.time_ns()
        self._setup()
        start_time_ns = time.time_ns()
        self._run()
        end_time_ns = time.time_ns()
        print(
            "# INITIALIZATION TIME " + str((start_time_ns - setup_start_time_ns) / 1e9)
        )
        print("# EXECUTION TIME " + str((end_time_ns - start_time_ns) / 1e9))

    def class_register(self, cls: Type) -> None:
        """register class. This method is used for user-defined classes.

        Usually, user-defined classes, i.e., the classes you implemented for your original simulation, cannot be referred from
        pams package, especially from simulation runners. Therefore, the class registration to the runner is necessary.

        Args:
            cls (Type): class to register.

        Returns:
            None
        """
        self.registered_class.append(cls)

    @abstractmethod
    def _setup(self) -> None:
        """internal usage class for setup. This method should be implemented in descendants."""
        pass

    @abstractmethod
    def _run(self) -> None:
        """internal usage class for simulation running. This method should be implemented in descendants.

        Usually the process in this methods should be control simulation flow and parallelization.
        """
        pass



from pams import agents
from pams import events
from pams import runners
from pams import utils
from pams.fundamentals import Fundamentals
from pams.index_market import IndexMarket
from pams.market import Market
from pams.order import LIMIT_ORDER
from pams.order import MARKET_ORDER
from pams.order import Cancel
from pams.order import Order
from pams.order import OrderKind
from pams.order_book import OrderBook
from pams.session import Session
from pams.simulator import Simulator
from pams.version import __version__



from typing import List
from typing import Optional

from pams.order import OrderKind


class Log:
    """Log class."""

    def read_and_write(self, logger: "Logger") -> None:
        """writing a log. (The actual writing timing depends on the logger.)

        Args:
            logger (:class:`pams.logs.Logger`): logger.

        Returns:
            None
        """
        logger.write(log=self)

    def read_and_write_with_direct_process(self, logger: "Logger") -> None:
        """direct writing a log. (This method is intended to call logger to write the log immediately.)

        Args:
            logger (:class:`pams.logs.Logger`): logger.

        Returns:
            None
        """
        logger.write_and_direct_process(log=self)


class OrderLog(Log):
    """Order log class.

    This log is usually generated when an order is accepted by markets.
    """

    def __init__(
        self,
        order_id: int,
        market_id: int,
        time: int,
        agent_id: int,
        is_buy: bool,
        kind: OrderKind,
        volume: int,
        price: Optional[float] = None,
        ttl: Optional[int] = None,
    ):
        """initialize.

        Args:
            order_id (int): order ID.
            market_id (int): market ID.
            time (int): time.
            agent_id (int): agent ID.
            is_buy (bool): whether it is a buy order or not.
            kind (:class:`pams.order.OrderKind`): kind of order.
            volume (int): order volume.
            price (float, Optional): order price.
            ttl (int, Optional): time to order expiration.
        """
        self.order_id: int = order_id
        self.market_id: int = market_id
        self.time: int = time
        self.agent_id: int = agent_id
        self.is_buy: bool = is_buy
        self.kind: OrderKind = kind
        self.price: Optional[float] = price
        self.volume: int = volume
        self.ttl: Optional[int] = ttl


class CancelLog(Log):
    """Cancel type log class.

    This log is usually generated when a cancel order is accepted by markets.
    """

    # ToDo: check ttl is necessary or not
    def __init__(
        self,
        order_id: int,
        market_id: int,
        cancel_time: int,
        order_time: int,
        agent_id: int,
        is_buy: bool,
        kind: OrderKind,
        volume: int,
        price: Optional[float] = None,
        ttl: Optional[int] = None,
    ):
        """initialize.

        Args:
            order_id (int): order ID.
            market_id (int): market ID.
            cancel_time (int): time to cancel.
            order_time (int): time to order.
            agent_id (int): agent ID.
            is_buy (bool): whether it is a buy order or not.
            kind (:class:`pams.order.OrderKind`): kind of order.
            volume (int): order volume.
            price (float, Optional): order price.
            ttl (int, Optional): time to cancel expiration.
        """
        self.order_id: int = order_id
        self.market_id: int = market_id
        self.cancel_time: int = cancel_time
        self.order_time: int = order_time
        self.agent_id: int = agent_id
        self.is_buy: bool = is_buy
        self.kind: OrderKind = kind
        self.price: Optional[float] = price
        self.volume: int = volume
        self.ttl: Optional[int] = ttl


class ExecutionLog(Log):
    """Execution type log class.

    This log is usually generated when an order is executed on markets.
    """

    def __init__(
        self,
        market_id: int,
        time: int,
        buy_agent_id: int,
        sell_agent_id: int,
        buy_order_id: int,
        sell_order_id: int,
        price: float,
        volume: int,
    ):
        """initialize.

        Args:
            market_id (int): market ID.
            time (int): time to execute.
            buy_agent_id (int): buyer agent ID.
            sell_agent_id (int): seller agent ID.
            buy_order_id (int): buy order ID.
            sell_order_id (int): sell order ID.
            price (float): executed price.
            volume (int): executed volume.
        """
        self.market_id: int = market_id
        self.time: int = time
        self.buy_agent_id: int = buy_agent_id
        self.sell_agent_id: int = sell_agent_id
        self.buy_order_id: int = buy_order_id
        self.sell_order_id: int = sell_order_id
        self.price: float = price
        self.volume: int = volume


class SimulationBeginLog(Log):
    """Simulation beginning log class."""

    def __init__(self, simulator: "Simulator"):  # type: ignore
        """initialize.

        Args:
            simulator (:class:`pams.Simulator`): simulator.
        """
        self.simulator = simulator


class SimulationEndLog(Log):
    """Simulation ending log class."""

    def __init__(self, simulator: "Simulator"):  # type: ignore
        """initialize.

        Args:
            simulator (:class:`pams.Simulator`): simulator.
        """
        self.simulator = simulator


class SessionBeginLog(Log):
    """Session beginning log class."""

    def __init__(self, session: "Session", simulator: "Simulator"):  # type: ignore
        """initialize.

        Args:
            session (:class:`pams.Session`): session.
            simulator (:class:`pams.Simulator`): simulator.
        """
        self.simulator = simulator
        self.session = session


class SessionEndLog(Log):
    """Session ending log class."""

    def __init__(self, session: "Session", simulator: "Simulator"):  # type: ignore
        """initialize.

        Args:
            session (:class:`pams.Session`): session.
            simulator (:class:`pams.Simulator`): simulator.
        """
        self.simulator = simulator
        self.session = session


class MarketStepBeginLog(Log):
    """Market step beginning log class."""

    def __init__(self, session: "Session", market: "Market", simulator: "Simulator"):  # type: ignore
        """initialize.

        Args:
            session (:class:`pams.Session`): session.
            market (:class:`pams.Market`): market.
            simulator (:class:`pams.Simulator`): simulator.
        """
        self.session = session
        self.market = market
        self.simulator = simulator


class MarketStepEndLog(Log):
    """Market step ending log class."""

    def __init__(self, session: "Session", market: "Market", simulator: "Simulator"):  # type: ignore
        """initialize.

        Args:
            session (:class:`pams.Session`): session.
            market (:class:`pams.Market`): market.
            simulator (:class:`pams.Simulator`): simulator.
        """
        self.session = session
        self.market = market
        self.simulator = simulator


class Logger:
    """Logger class.

    Logger is designed to handling parallelized market simulations.
    In the parallelized simulation, the order that logs are pushed to this logger is not deterministic
     because of varied computational time.
    For example, there are 2 markets, market 1 and market 2, and those markets are run in parallel,
     the computational time of one step for each market are not fixed.
    Therefore, the logging sequence of market 1 and market 2 is not always the same.
    For handling this problem, logger should save the logs from those markets and the end of each step,
     the logs are processed and written.
    However, for some implementation, non-blocking log writing should be also supported.
    Therefore, this logger has 2 methods to handling logs -- :func:`write` and :func:`write_and_direct_process`
    """

    simulator: "Simulator"  # type: ignore

    def __init__(self) -> None:
        """initialize logger."""
        self.pending_logs: List[Log] = []

    def _set_simulator(self, simulator: "Simulator") -> None:  # type: ignore
        """set simulator.

        Args:
            simulator (:class:`pams.Simulator`): simulator.

        Returns:
            None
        """
        self.simulator = simulator

    def write(self, log: "Log") -> None:
        """set a log to pending list.

        Args:
            log (:class:`pams.logs.Log`): log.

        Returns:
            None
        """
        self.pending_logs.append(log)

    def bulk_write(self, logs: List["Log"]) -> None:
        """set some logs to pending list.

        Args:
            logs (List[:class:`pams.logs.Log`]): log list.

        Returns:
            None
        """
        self.pending_logs.extend(logs)

    def write_and_direct_process(self, log: "Log") -> None:
        """direct writing a log.

        Args:
            log (:class:`pams.logs.Log`): log.

        Returns:
            None
        """
        self.process(logs=[log])

    def bulk_write_and_direct_process(self, logs: List["Log"]) -> None:
        """direct writing some logs.

        Args:
            logs (List[:class:`pams.logs.Log`]): log list.

        Returns:
            None
        """
        self.process(logs=logs)

    def _process(self) -> None:
        self.process(logs=self.pending_logs)
        self.pending_logs = []

    def process(self, logs: List["Log"]) -> None:
        """logging execution. For each log type, each processing method are implemented.

        For usual implementation, please use

         - :func:`process_order_log`
         - :func:`process_cancel_log`
         - :func:`process_execution_log`
         - :func:`process_simulation_begin_log`
         - :func:`process_simulation_end_log`
         - :func:`process_session_begin_log`
         - :func:`process_session_end_log`
         - :func:`process_market_step_begin_log`
         - :func:`process_market_step_end_log`

        However, if you want to control the log writing sequence, you can change this implementation.

        Args:
            logs (List[:class:`pams.logs.Log`]): log list.

        Returns:
            None
        """
        for log in logs:
            if isinstance(log, OrderLog):
                self.process_order_log(log=log)
            elif isinstance(log, CancelLog):
                self.process_cancel_log(log=log)
            elif isinstance(log, ExecutionLog):
                self.process_execution_log(log=log)
            elif isinstance(log, SimulationBeginLog):
                self.process_simulation_begin_log(log=log)
            elif isinstance(log, SimulationEndLog):
                self.process_simulation_end_log(log=log)
            elif isinstance(log, SessionBeginLog):
                self.process_session_begin_log(log=log)
            elif isinstance(log, SessionEndLog):
                self.process_session_end_log(log=log)
            elif isinstance(log, MarketStepBeginLog):
                self.process_market_step_begin_log(log=log)
            elif isinstance(log, MarketStepEndLog):
                self.process_market_step_end_log(log=log)
            else:
                raise NotImplementedError

    def process_order_log(self, log: "OrderLog") -> None:
        """process order log. Called from :func:`process`.

        Args:
            log (:class:`pams.logs.OrderLog`]): order log

        Returns:
            None"""
        pass

    def process_cancel_log(self, log: "CancelLog") -> None:
        """process cancel log. Called from :func:`process`.

        Args:
            log (:class:`pams.logs.CancelLog`]): cancel log

        Returns:
            None
        """
        pass

    def process_execution_log(self, log: "ExecutionLog") -> None:
        """process execution log. Called from :func:`process`.

        Args:
            log (:class:`pams.logs.ExecutionLog`]): execution log

        Returns:
            None
        """
        pass

    def process_simulation_begin_log(self, log: "SimulationBeginLog") -> None:
        """process simulation begin log. Called from :func:`process`.

        Args:
            log (:class:`pams.logs.SimulationBeginLog`]): simulation begin log

        Returns:
            None
        """
        pass

    def process_simulation_end_log(self, log: "SimulationEndLog") -> None:
        """process simulation end log. Called from :func:`process`.

        Args:
            log (:class:`pams.logs.SimulationEndLog`]): simulation end log

        Returns:
            None
        """
        pass

    def process_session_begin_log(self, log: "SessionBeginLog") -> None:
        """process session begin log. Called from :func:`process`.

        Args:
            log (:class:`pams.logs.SessionBeginLog`]): session begin log

        Returns:
            None
        """
        pass

    def process_session_end_log(self, log: "SessionEndLog") -> None:
        """process session end log. Called from :func:`process`.

        Args:
            log (:class:`pams.logs.SessionEndLog`]): session end log

        Returns:
            None
        """
        pass

    def process_market_step_begin_log(self, log: "MarketStepBeginLog") -> None:
        """process market step begin log. Called from :func:`process`.

        Args:
            log (:class:`pams.logs.MarketStepBeginLog`]): market step begin log

        Returns:
            None
        """
        pass

    def process_market_step_end_log(self, log: "MarketStepEndLog") -> None:
        """process market step end log. Called from :func:`process`.

        Args:
            log (:class:`pams.logs.MarketStepEndLog`]): market step end log

        Returns:
            None
        """
        pass



from .base import CancelLog
from .base import ExecutionLog
from .base import Log
from .base import Logger
from .base import MarketStepBeginLog
from .base import MarketStepEndLog
from .base import OrderLog
from .base import SessionBeginLog
from .base import SessionEndLog
from .base import SimulationBeginLog
from .base import SimulationEndLog
from .market_step_loggers import MarketStepPrintLogger
from .market_step_loggers import MarketStepSaver



from typing import List
from typing import Union

from ..market import Market
from ..order import LIMIT_ORDER
from ..order import Cancel
from ..order import Order
from .base import Agent


class TestAgent(Agent):
    """TestAgent class

    This agent is mainly used for generating orders with zero intelligence and just aims for market tests.
    Different from Agent class, this agent is implementing its `submit_orders` methods.

    Note:
        Please do not use this agent for your research.
    """

    def submit_orders(self, markets: List[Market]) -> List[Union[Order, Cancel]]:
        """submit orders. This method is usually called by simulator or runners.

        Args:
             markets (list[Market]): market objects. Usually it is set by simulator or runners automatically.

        Returns:
            list[Order, Cancel]: orders including cancel orders.

        Note:
            TestAgent is mainly introduced for testing.
            Therefore, we do not intend that you modify it.
            If you want to use similar agents for your research purpose, please make user-defined class instead of using TestAgent.
        """
        margin_scale: float = 10.0
        volume_scale: int = 100
        time_length_scale: int = 100
        buy_chance: float = 0.4
        sell_chance: float = 0.4

        orders: List[Union[Order, Cancel]] = []
        for market in markets:
            if self.is_market_accessible(market_id=market.market_id):
                price: float = market.get_market_price() + (
                    self.prng.random() * 2 * margin_scale - margin_scale
                )
                volume: int = self.prng.randint(1, volume_scale)
                time_length: int = self.prng.randint(1, time_length_scale)
                p = self.prng.random()
                if p < buy_chance + sell_chance:
                    orders.append(
                        Order(
                            agent_id=self.agent_id,
                            market_id=market.market_id,
                            is_buy=p < buy_chance,
                            kind=LIMIT_ORDER,
                            volume=volume,
                            price=price,
                            ttl=time_length,
                        )
                    )
        return orders



from typing import Dict
from typing import List

from .base import Logger
from .base import MarketStepEndLog


class MarketStepPrintLogger(Logger):
    """Logger of the market step class."""

    def process_market_step_end_log(self, log: MarketStepEndLog) -> None:
        """print the market log.

        Args:
            log (:class:`pams.logs.MarketStepEndLog`): matket log.

        Returns:
            None
        """
        print(
            f"{log.session.session_id} {log.market.get_time()} {log.market.market_id} {log.market.name} {log.market.get_market_price()} {log.market.get_fundamental_price()}"
        )


class MarketStepSaver(Logger):
    """Saver of the market step class."""

    market_step_logs: List[Dict] = []

    def process_market_step_end_log(self, log: MarketStepEndLog) -> None:
        """stack the market log.

        Args:
            log (:class:`pams.logs.MarketStepEndLog`): market log.

        Returns:
            None
        """
        self.market_step_logs.append(
            {
                "session_id": log.session.session_id,
                "market_time": log.market.get_time(),
                "market_id": log.market.market_id,
                "market_name": log.market.name,
                "market_price": log.market.get_market_price(),
                "fundamental_price": log.market.get_fundamental_price(),
            }
        )



import warnings
from dataclasses import dataclass
from typing import Optional
from typing import cast


@dataclass(frozen=True)
class OrderKind:
    kind_id: int
    name: str

    def __repr__(self) -> str:
        return self.name

    def __eq__(self, other: object) -> bool:
        if other.__class__ != self.__class__:
            return False
        other = cast(OrderKind, other)
        return other.kind_id == self.kind_id

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return self.kind_id


MARKET_ORDER = OrderKind(kind_id=0, name="MARKET_ORDER")
LIMIT_ORDER = OrderKind(kind_id=1, name="LIMIT_ORDER")


class Order:
    def __init__(
        self,
        agent_id: int,
        market_id: int,
        is_buy: bool,
        kind: OrderKind,
        volume: int,
        placed_at: Optional[int] = None,
        price: Optional[float] = None,
        order_id: Optional[int] = None,
        ttl: Optional[int] = None,
    ):
        if kind == MARKET_ORDER and price is not None:
            raise ValueError("price have to be None when kind is MARKET_ORDER")
        if kind == LIMIT_ORDER and price is None:
            raise ValueError("price have to be set when kind is LIMIT_ORDER")
        if price is not None and price <= 0:
            warnings.warn("price should be positive")
        if volume <= 0:
            raise ValueError("volume have to be positive")
        if ttl is not None and ttl <= 0:
            raise ValueError("ttl have to be positive or None")

        self.agent_id: int = agent_id
        self.market_id: int = market_id
        self.is_buy: bool = is_buy
        self.kind: OrderKind = kind
        self.volume: int = volume
        self.placed_at: Optional[int] = placed_at
        self.price: Optional[float] = price
        self.order_id: Optional[int] = order_id
        self.ttl: Optional[int] = ttl
        self.is_canceled: bool = False

    def check_system_acceptable(self, agent_id: int) -> None:
        if agent_id != self.agent_id:
            raise AttributeError("agent_id is fake")
        if self.placed_at is not None:
            raise AttributeError(
                "this order is already submitted to a market or placed_at have to be set to None"
            )
        if self.is_canceled is True:
            raise AttributeError("this order is already canceled")

    def is_expired(self, time: int) -> bool:
        if self.placed_at is None:
            raise Exception("this order is not yet placed to a market")
        if self.ttl is None:
            return False
        else:
            return self.placed_at + self.ttl < time

    def extra_repr(self) -> str:
        return (
            f"id={self.order_id}, kind={self.kind}, is_buy={self.is_buy}, price={self.price}, volume={self.volume}, "
            + f"agent={self.agent_id}, market={self.market_id}, placed_at={self.placed_at}, ttl={self.ttl}, "
            + f"is_canceled={self.is_canceled}"
        )

    def __gt__(self, other: object) -> bool:
        if self.__class__ != other.__class__:
            raise NotImplementedError(
                f"not supporting the comparison between Order and {other.__class__}"
            )
        other = cast(Order, other)
        if self.is_buy != other.is_buy:
            raise ValueError(
                "not supporting the comparison between sell and buy orders"
            )

        def _compare_placed_at(a: Order, b: Order) -> bool:
            if a.placed_at is None and b.placed_at is None:
                raise ValueError("orders still not placed cannot be compared")
            elif a.placed_at is None:
                return True
            elif b.placed_at is None:
                return False
            else:
                return a.placed_at > b.placed_at

        if self.kind == MARKET_ORDER and other.kind == MARKET_ORDER:
            return _compare_placed_at(a=self, b=other)
        elif self.kind == MARKET_ORDER:
            return False
        elif other.kind == MARKET_ORDER:
            return True
        else:
            if self.kind != LIMIT_ORDER or other.kind != LIMIT_ORDER:
                raise NotImplementedError
            else:
                if self.price is None or other.price is None:
                    raise AssertionError
                if self.price != other.price:
                    if self.is_buy:
                        return self.price < other.price
                    else:
                        return self.price > other.price
                else:
                    return _compare_placed_at(a=self, b=other)


class Cancel:
    def __init__(self, order: Order, placed_at: Optional[int] = None):
        self.order: Order = order
        self.placed_at: Optional[int] = placed_at

    @property
    def agent_id(self) -> int:
        return self.order.agent_id

    @property
    def market_id(self) -> int:
        return self.order.market_id

    def check_system_acceptable(self, agent_id: int) -> None:
        if agent_id != self.order.agent_id:
            raise AttributeError("canceling other's order")
        if self.placed_at is not None:
            raise AttributeError(
                "this cancel order is already submitted to a market or placed_at have to be set to None"
            )
        if self.order.is_canceled is True:
            raise AttributeError("this order is already canceled")



import random
from abc import ABC
from abc import abstractmethod
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from ..logs.base import CancelLog
from ..logs.base import ExecutionLog
from ..logs.base import Logger
from ..logs.base import OrderLog
from ..market import Market
from ..order import Cancel
from ..order import Order
from ..utils.json_random import JsonRandom


class Agent(ABC):
    """Agent class (abstract class)

    Once you define the agent class inheriting this agent ABC class, simulator automatically generate agents using the class you define.
    Therefore, you don't need to call __init__ or any other method and simulator automatically call them if it is necessary.

    `submit_orders(self, markets: List[Market])` is required to be implemented.

    .. seealso::
        - :class:`pams.agents.FCNAgent`: FCNAgent
    """

    def __init__(
        self,
        agent_id: int,
        prng: random.Random,
        simulator: "Simulator",  # type: ignore
        name: str,
        logger: Optional[Logger] = None,
    ) -> None:
        """agent initialization. Usually call from simulator automatically.

        Args:
            agent_id (int): agent id. This is also required to generate orders to identify who submit orders.
            prng (random.Random): pseudo random number generator for this agent.
            simulator (Simulator): this is used for accessing simulation environment.
            name (str): this is unique name for this agent.
                        Automatically generated by simulator and just set it as the name of this agent.
            logger (Logger, Optional): logger for correcting various outputs in one simulation.
                                       logger is usually shared to all classes.
                                       Please note that logger is usually not thread-safe and non-blocking.

        Returns:
            None

        Note:
             `prng` should not be shared with other classes and be used only in this class.
             It is because sometimes agent process runs one of parallelized threads.
        """
        self.agent_id: int = agent_id
        self.name: str = name
        self.asset_volumes: Dict[int, int] = {}
        self.cash_amount: float = 0
        self.prng: random.Random = prng
        self.sim: "Simulator" = simulator  # type: ignore
        self.logger: Optional[Logger] = logger

    def setup(self, settings: Dict[str, Any], accessible_markets_ids: List[int], *args, **kwargs) -> None:  # type: ignore
        """agent setup. Usually be called from simulator/runner automatically.

        Args:
            settings (Dict[str, Any]): agent configuration.  Usually, automatically set from json config of simulator.
                                       This must include the parameters "cashAmount" and "assetVolume".
            accessible_markets_ids (List[int]): list of market IDs.

        Returns:
            None
        """
        if "cashAmount" not in settings:
            raise ValueError("cashAmount is required property of agent settings")
        self.cash_amount = settings["cashAmount"]
        if "assetVolume" not in settings:
            raise ValueError("cashAmount is required property of agent settings")

        for market_id in accessible_markets_ids:
            self.set_market_accessible(market_id=market_id)
            volume = int(
                JsonRandom(prng=self.prng).random(json_value=settings["assetVolume"])
            )
            self.set_asset_volume(market_id=market_id, volume=volume)

    def get_asset_volume(self, market_id: int) -> int:
        """getter of the asset volume held by the agent.

        Args:
            market_id (int): market ID.

        Returns:
            int: asset volume for the specified market ID.
        """
        if not self.is_market_accessible(market_id=market_id):
            raise AssertionError(f"market {market_id} is not accessible")
        return self.asset_volumes[market_id]

    def get_cash_amount(self) -> float:
        """getter of the cash amount held by the agent.

        Returns:
            float: cash amount held by this agent.
        """
        return self.cash_amount

    def get_prng(self) -> random.Random:
        """getter of the pseudo random number generator.

        Returns:
            random.Random: pseudo random number generator for this agent.
        """
        return self.prng

    def is_market_accessible(self, market_id: int) -> bool:
        """determine if the market ID is included in the asset volume

        Args:
            market_id (int): market ID.

        Returns:
            bool: whether the asset volume held by the agent contains the specified market or not.
        """
        return market_id in self.asset_volumes

    def submitted_order(self, log: OrderLog) -> None:
        """call back when an order submission is accepted by a market.

        Args:
            log (OrderLog): log for order submission

        Returns:
            None
        """
        pass

    def executed_order(self, log: ExecutionLog) -> None:
        """call back when a submitted order is executed in a market.

        Args:
            log (ExecutionLog): log for order execution

        Returns:
            None
        """
        pass

    def canceled_order(self, log: CancelLog) -> None:
        """call back when cancel order is accepted by a market.

        Args:
            log (CancelLog): log for order cancellation

        Returns:
            None
        """
        pass

    def set_asset_volume(self, market_id: int, volume: int) -> None:
        """setter of the asset volume held by agent.

        Args:
            market_id (int): market ID.
            volume (int): volume to be set for the market

        Returns:
            None
        """
        if not self.is_market_accessible(market_id=market_id):
            raise AssertionError(f"market {market_id} is not accessible")
        self.asset_volumes[market_id] = volume

    def set_cash_amount(self, cash_amount: float) -> None:
        """setter of the cash amount held by agent.

        Args:
            cash_amount (float): cash amount held by this agent.

        Returns:
            None
        """
        self.cash_amount = cash_amount

    def set_market_accessible(self, market_id: int) -> None:
        """set the specified market volume to 0.

        Args:
            market_id (int): market ID.

        Returns:
            None
        """
        self.asset_volumes[market_id] = 0

    @abstractmethod
    def submit_orders(self, markets: List[Market]) -> List[Union[Order, Cancel]]:
        """submit orders (abstract method). This method automatically called from runners.

        This method is called only when this agent has a chance to submit orders.
        Therefore, it is not guaranteed that this method is called at all the step of simulation.

        Args:
            markets (List[Market]): markets to order.

        Returns:
            List[Union[Order, Cancel]]: order list.

        Note:
            You should implement this method if you inherit this agent.
        """
        pass

    def update_asset_volume(self, market_id: int, delta: int) -> None:
        """increasing or decreasing the asset volume.

        Args:
            market_id (int): market ID.
            delta (int): amount of change in the asset volume.

        Returns:
            None
        """
        if not self.is_market_accessible(market_id=market_id):
            raise AssertionError(f"market {market_id} is not accessible")
        self.asset_volumes[market_id] += delta

    def update_cash_amount(self, delta: float) -> None:
        """increasing or decreasing the cash amount.

        Args:
            delta (float): amount of change in the cash amount.

        Returns:
            None
        """
        self.cash_amount += delta

    def __repr__(self) -> str:
        """string representation of agent class.

        Returns:
            str: string representation of this class.
        """
        return (
            self.__class__.__name__
            + str(id)
            + ","
            + str(self.cash_amount)
            + ","
            + str(self.asset_volumes)
        )



from typing import Any
from typing import Dict
from typing import List

from .base import EventABC
from .base import EventHook


class FundamentalPriceShock(EventABC):
    """This suddenly changes the fundamental price (just changing it).

    This event is only called via :func:`hooked_before_step_for_market` at designated step.
    """

    target_market_name: str
    target_market: "Market"  # type: ignore
    trigger_time: int
    price_change_rate: float
    is_enabled: bool = True
    shock_time_length: int = 1

    def setup(self, settings: Dict[str, Any], *args, **kwargs) -> None:  # type: ignore
        """event setup. Usually be called from simulator/runner automatically.

        Args:
            settings (Dict[str, Any]): agent configuration. Usually, automatically set from json config of simulator.
                                       This must include the parameters "triggerDays", "target", "triggerTime", and "priceChangeRate".
                                       This can include the parameters "enabled" and "shockTimeLength".

        Returns:
            None
        """
        if "triggerDays" in settings:
            raise ValueError("triggerDays and numStepsOneDay are obsoleted.")
        if "target" not in settings:
            raise ValueError("target is required for FundamentalPriceShock")
        self.target_market_name = settings["target"]
        if "triggerTime" not in settings:
            raise ValueError("triggerTime is required for FundamentalPriceShock")
        self.trigger_time = self.session.session_start_time + settings["triggerTime"]
        if "priceChangeRate" not in settings:
            raise ValueError("priceChangeRate is required for FundamentalPriceShock")
        self.price_change_rate = settings["priceChangeRate"]
        if "enabled" in settings:
            self.is_enabled = settings["enabled"]
        if "shockTimeLength" in settings:
            self.shock_time_length = settings["shockTimeLength"]
        self.target_market = self.simulator.name2market[self.target_market_name]

    def hook_registration(self) -> List[EventHook]:
        event_hook = EventHook(
            event=self,
            hook_type="market",
            is_before=True,
            time=[self.trigger_time + i for i in range(self.shock_time_length)],
            specific_instance=self.target_market,
        )
        return [event_hook]

    def hooked_before_step_for_market(self, simulator: "Simulator", market: "Market") -> None:  # type: ignore
        time: int = market.get_time()
        if not (self.trigger_time <= time < self.trigger_time + self.shock_time_length):
            raise AssertionError
        if market != self.target_market:
            raise AssertionError
        market.change_fundamental_price(scale=1 + self.price_change_rate)


FundamentalPriceShock.hook_registration.__doc__ = EventABC.hook_registration.__doc__
FundamentalPriceShock.hooked_before_step_for_market.__doc__ = (
    EventABC.hooked_before_step_for_market.__doc__
)



import random
from abc import ABC
from abc import abstractmethod
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Type


class EventHook:
    """Event Hook class.

    Event hook define when and what events are hooked from simulator.
    It means that Event ( :class:`Event` ) can be used for multiple time if you appropriately set the event hook using the evnet.
    Event can be hooked before and after order placements, order cancellations, order executions, each session, and each step mor markets.
    You can also filter hooking point by market time, classes, instances of markets. (Currently, only market class and instances are supported.)

    .. seealso:
        - :class:FundamentalPriceShock
    """

    def __init__(
        self,
        event: "EventABC",
        hook_type: str,
        is_before: bool,
        time: Optional[List[int]] = None,
        specific_class: Optional[Type] = None,
        specific_instance: Optional[object] = None,
    ):
        """Event Hook initialization.

        Args:
            event (:class:`pams.events.EventABC`): event instance.
            hook_type (str): hook type. This must be "order", "cancel", "execution", "session", or "market".
            is_before (bool): flag whether to run before or not. If hook_type is "execution", this must be set to False.
            time (List[int], Optional): event execution time.
            specific_class (Type, Optional): specific class. If this is specified, the hook_type must be "market".
                                             (In future, it could be expanded to other types.)
            specific_instance (object, Optional): specific instance. If this is specified, the hook_type must be "market".
                                                  (In future, it could be expanded to other types.)

        Returns:
            None
        """
        if hook_type not in ["order", "cancel", "execution", "session", "market"]:
            raise ValueError(
                "hook type have to be order, execution, session, or market"
            )
        if hook_type == "execution" and is_before:
            raise ValueError("execution can be hooked only after it")
        self.event = event
        self.hook_type = hook_type
        self.is_before = is_before
        self.time = time
        if specific_class is not None or specific_instance is not None:
            if hook_type not in ["market"]:
                raise ValueError(
                    "specific_class and specific_instance are not supported except for market"
                )
        self.specific_class: Optional[Type] = (
            specific_class.__class__ if specific_class is not None else None
        )
        self.specific_instance: Optional[object] = specific_instance


class EventABC(ABC):
    """event base class (ABC class).

    It defines what the event is.
    All events should inherit this class.

    .. seealso::
        - :class:`pams.events.FundamentalPriceShock`
    """

    def __init__(
        self,
        event_id: int,
        prng: random.Random,
        session: "Session",  # type: ignore
        simulator: "Simulator",  # type: ignore
        name: str,
    ) -> None:
        """event initialization. Usually be called from simulator/runner automatically.

        Args:
            event_id (int): event ID.
            prng (random.Random): pseudo random number generator for this event.
            session (:class:`pams.Session`): session to execute this event.
            simulator (:class:`pams.Simulator`): simulator that executes this event.
            name (str): event name.

        Returns:
            None
        """
        self.event_id: int = event_id
        self.prng: random.Random = prng
        self.simulator = simulator
        self.name: str = name
        self.session = session

    def setup(self, settings: Dict[str, Any], *args, **kwargs) -> None:  # type: ignore
        """event setup. Usually be called from simulator/runner automatically.

        Args:
            settings (Dict[str, Any]): agent configuration. Usually, automatically set from json config of simulator.

        Returns:
            None
        """
        pass

    @abstractmethod
    def hook_registration(self) -> List[EventHook]:
        """Define when this event should be hooked by simulator.
        This method is automatically hooked by simulator at the beginning of simulation.
        You must implement this.

        Returns:
            List[EventHook]: The list of event hook ( :class:`EventHook` )
        """
        pass

    def hooked_before_order(self, simulator: "Simulator", order: "Order") -> None:  # type: ignore
        """This method is hooked before order placements if you set the event hook.
        Please be careful that the order haven't yet been accepted by markets and it could be leakage of order information.

        .. seealso:
            - ToDo: simulation flow

        Args:
            simulator (Simulator): simulator for reference.
            order (Order): order accepting now.
        """
        pass

    def hooked_after_order(self, simulator: "Simulator", order_log: "OrderLog") -> None:  # type: ignore
        """This method is hooked after order placements if you set the event hook.
        Please be careful that the order haven't yet been executed if it could be executed immediately.

        .. seealso:
            - ToDo: simulation flow

        Args:
            simulator (Simulator): simulator for reference.
            order_log (OrderLog): order accepted.
        """
        pass

    def hooked_before_cancel(self, simulator: "Simulator", cancel: "Cancel") -> None:  # type: ignore
        """This method is hooked before order cancellations if you set the event hook.
        Please be careful that the cancel order haven't yet been executed.

        .. seealso:
            - ToDo: simulation flow

        Args:
            simulator (Simulator): simulator for reference.
            cancel (Cancel): cancel order submitted.
        """
        pass

    def hooked_after_cancel(self, simulator: "Simulator", cancel_log: "CancelLog") -> None:  # type: ignore
        """This method is hooked after order cancellations if you set the event hook.

        .. seealso:
            - ToDo: simulation flow

        Args:
            simulator (Simulator): simulator for reference.
            cancel_log (CancelLog): cancel order submitted.
        """
        pass

    def hooked_after_execution(self, simulator: "Simulator", execution_log: "ExecutionLog") -> None:  # type: ignore
        """This method is hooked after order executions if you set the event hook.

        .. seealso:
            - ToDo: simulation flow

        Args:
            simulator (Simulator): simulator for reference.
            execution_log (ExecutionLog): execution log.
        """
        pass

    def hooked_before_session(self, simulator: "Simulator", session: "Session") -> None:  # type: ignore
        """This method is hooked before session beginnings if you set the event hook.

        .. seealso:
            - ToDo: simulation flow

        Args:
            simulator (Simulator): simulator for reference.
            session (Session): session to be started.
        """
        pass

    def hooked_after_session(self, simulator: "Simulator", session: "Session") -> None:  # type: ignore
        """This method is hooked after session ends if you set the event hook.

        .. seealso:
            - ToDo: simulation flow

        Args:
            simulator (Simulator): simulator for reference.
            session (Session): session to be ended.
        """
        pass

    def hooked_before_step_for_market(self, simulator: "Simulator", market: "Market") -> None:  # type: ignore
        """This method is hooked at each step before market processing if you set the event hook.

        .. seealso:
            - ToDo: simulation flow

        Args:
            simulator (Simulator): simulator for reference.
            market (Market): market to be processed.
        """
        pass

    def hooked_after_step_for_market(self, simulator: "Simulator", market: "Market") -> None:  # type: ignore
        """This method is hooked at each step after market processing if you set the event hook.

        .. seealso:
            - ToDo: simulation flow

        Args:
            simulator (Simulator): simulator for reference.
            market (Market): market processed.
        """
        pass
